/* program : pc2key.c
   author  : Ren Bodenstaff
   date    : august 14 1994
   purpose : swaps member groups for key lock
   tested  : <none>
   nb      : compile this program with Test Stack Overflow OFF
*/

/* ---------------- Includes ------------------------------ */
#include <time.h>
#include <stdlib.h>
#include <io.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "serial.h"

/* ---------------- Defines ------------------------------- */
#define VERSION 0x00

#define FALSE           0
#define TRUE           (!FALSE)

#define NOERROR         0       /* no error               */
#define BUFOVFL         1       /* buffer overflowed      */

#define ESC             0x1B    /* ascii escape character */
#define ASCII           0x007F  /* mask ASCII characters  */
#define SBUFSIZ         0x4000  /* serial buffer size     */

#define MAXLINE         80
#define ID_LEN          7       /* length of member number + '\0' in bytes */
#define MAX_MEM         680     /* maximum number of members per lock      */
#define LOCK_LEN	9	/* length of lock id in bytes + '\0' */

/* ---------------- Prototypes ---------------------------- */

/* --- own functions */
void scan_member(char *str, char *mem);
int read_names(char *list[], FILE *fp);
int read_lock(char *list[], int port);
void dump_list(char *list[], int count);
void write_list(char *list[], int count);
int create_union_list(char *cur_list[], char *mem_list[], char *uni_list[],
		      int cur_count, int mem_count);
int create_add_list(char *mem_list[], char *uni_list[], char *add_list[],
		    int mem_count, int uni_count);
int create_del_list(char *cur_list[], char *uni_list[], char *del_list[],
		    int cur_count, int uni_count);
int is_element(char *str, char *list[], int count);
void del_key(char *list[], int count);
void add_key(char *list[], int count);

/* --- serial functions */
void interrupt com_int(void);
int SerialOut(char x);
int SerialOutTough(char x);
void SerialString(char *string);
int getccb(void);
void setvects(void);
void resvects(void);
void i_enable(int pnum);
void i_disable(void);
void comm_on(void);
void comm_off(void);
void initserial(void);
void closeserial(void);
int SetPort(int Port);
int SetSpeed(int Speed);
int SetOthers(int Parity, int Bits, int StopBit);
int SetSerial(int Port, int Speed, int Parity, int Bits, int StopBit);
int c_break(void);

/* -------------------------- MAIN ---------------------------- */

/* --- some global variables */
  int SError = NOERROR;
  int portbase = 0;
  void interrupt(*oldvects[2])();

  static char ccbuf[SBUFSIZ];
  unsigned int startbuf = 0;
  unsigned int endbuf = 0;

main(int argc, char *argv[])
{
  FILE *names_file;
  int i, n, port;
  unsigned member_count;
  unsigned current_count;
  unsigned union_count;
  unsigned add_count;
  unsigned del_count;
  char *member_list[MAX_MEM];
  char *current_list[MAX_MEM];
  char *union_list[MAX_MEM];
  char *add_list[MAX_MEM];
  char *del_list[MAX_MEM];
  char ch;
  char line[MAXLINE];

  int c, done  = FALSE;

  /* --- check for command line consistency & open names file and scan port# */
  if(argc!=2 || (strcmp(argv[1], "com1") && strcmp(argv[1], "com2"))) {
    fprintf(stderr, "usage: key2pc <com1|com2>\n");
    exit(1);
  }
  if(strcmp(argv[1], "com1"))
    port=COM2;
  else
    port=COM1;

  fprintf(stdout, "current members of lock connected to com%d:\n", port);

  /* --- build lists of current and member numbers */
  current_count=read_lock(current_list, port);

  fprintf(stderr, "current number of members: %d\n", current_count);

  /* --- check for serial errors */
  switch (SError)
  {
    case NOERROR: closeserial();
                  
    case BUFOVFL: /* fprintf(stderr, "\nBuffer Overflow\n"); */
                  closeserial();
                  return (99);

    default:      fprintf(stderr, "\nUnknown Error, SError = %d\n", SError);
                  closeserial();
                  return (99);
  } /* switch */

  exit(0);
} /* main */

/* ---------------- Functions ---------------------------------- */

/* --- functions for list data structure */

int read_lock(char *list[], int port)
{
  /* --- read key members from lock by asserting 's' command
     --- and store current member numbers in lock_list
     --- nb: beware of master number (never delete this one)!!
  */

  char ch, line[MAXLINE], lock_id[LOCK_LEN];
  char *tmp;
  int i=0;
  int count=0;

  int speed    = 9600;
  int parity   = NO_PARITY;
  int bits     = 8;
  int stopbits = 1;

  /* --- first set communications parameters */

  if (SetSerial(port, speed, parity, bits, stopbits) != 0) {
    fprintf(stderr, "Serial Port setup error.\n");
    return (99);
  }

  initserial();
  ctrlbrk(c_break);

  /* --- send 's' command and collect data in list */

  SerialOutTough('s');

  i=0;
  do {
  if((ch=(char)getccb())!=-1) {
    if(ch!='\n')
      line[i++]=ch;
    else {
      if(!(i<ID_LEN)) { /* don't store non-valid lines */
	line[i-1]='\0';
	if(0) {
	  fprintf(stderr, "r_l: out of memory\n");
	  return(-1);
	} /* malloc */
	else {
	  fprintf(stdout, "%s\n", line);
	  count++;
	}
      }
      i=0;
    }
  } /* if((ch=(char)....) */
  } while (i<LOCK_LEN);

  return(count);
} /* read_lock */


/* --- functions for RS-232 communications (ripped from serial.c) */

/* Handle communications interrupts and put them in ccbuf */
void interrupt com_int(void)
{
  disable();
  if ((inportb(portbase + IIR) & RX_MASK) == RX_ID) {
    if (((endbuf + 1) & SBUFSIZ - 1) == startbuf)
      SError = BUFOVFL;

    ccbuf[endbuf++] = inportb(portbase + RXR);
    endbuf &= SBUFSIZ - 1;
  }

  /* Signal end of hardware interrupt */
  outportb(ICR, EOI);
  enable();
}

/* Output a character to the serial port */
int SerialOut(char x)
{
  long int timeout = 0x0000FFFFL;

  outportb(portbase + MCR,  MC_INT | DTR | RTS);

  /* Wait for Clear To Send from modem */
  while ((inportb(portbase + MSR) & CTS) == 0)
    if (!(--timeout))
      return (-1);

  timeout = 0x0000FFFFL;

  /* Wait for transmitter to clear */
  while ((inportb(portbase + LSR) & XMTRDY) == 0)
    if (!(--timeout))
      return (-1);

  disable();
  outportb(portbase + TXR, x);
  enable();

  return (0);
}

/* Output a character to the serial port with no hand shaking */
/* --- This hack is necessary, because the key lacks RTS & CTS --- */
int SerialOutTough(char x)
{
  outportb(portbase + MCR,  MC_INT | DTR | RTS);

  disable();
  outportb(portbase + TXR, x);
  enable();

  return (0);
}

/* Output a string to the serial port */
void SerialString(char *string)
{
  while (*string)
    SerialOut(*string++);
}

/* This routine returns the current value in the buffer */
int getccb(void)
{
  int res;

  if (endbuf == startbuf)
    return (-1);

  res = (int) ccbuf[startbuf++];
  startbuf %= SBUFSIZ;

  return (res);
}

/* Install our functions to handle communications */
void setvects(void)
{
  oldvects[0] = getvect(0x0B);
  oldvects[1] = getvect(0x0C);
  setvect(0x0B, com_int);
  setvect(0x0C, com_int);
}

/* Uninstall our vectors before exiting the program */
void resvects(void)
{
  setvect(0x0B, oldvects[0]);
  setvect(0x0C, oldvects[1]);
}

/* Turn on communications interrupts */
void i_enable(int pnum)
{
  int c;

  disable();
  c = inportb(portbase + MCR) | MC_INT;
  outportb(portbase + MCR, c);
  outportb(portbase + IER, RX_INT);
  c = inportb(IMR) & (pnum == COM1 ? IRQ4 : IRQ3);
  outportb(IMR, c);
  enable();
}

/* Turn off communications interrupts */
void i_disable(void)
{
  int c;

  disable();
  c = inportb(IMR) | ~IRQ3 | ~IRQ4;
  outportb(IMR, c);
  outportb(portbase + IER, 0);
  c = inportb(portbase + MCR) & ~MC_INT;
  outportb(portbase + MCR, c);
  enable();
}

/* Tell modem that we're ready to go */
void comm_on(void)
{
  int c, pnum;

  pnum = (portbase == COM1BASE ? COM1 : COM2);
  i_enable(pnum);
  c = inportb(portbase + MCR) | DTR | RTS;
  outportb(portbase + MCR, c);
}

/* Go off-line */
void comm_off(void)
{
  i_disable();
  outportb(portbase + MCR, 0);
}

void initserial(void)
{
  endbuf = startbuf = 0;
  setvects();
  comm_on();
}

void closeserial(void)
{
  comm_off();
  resvects();
}

/* Set the port number to use */
int SetPort(int Port)
{
  int  Offset, far *RS232_Addr;

  switch (Port)
  { /* Sort out the base address */
    case COM1 : Offset = 0x0000;
      break;
    case COM2 : Offset = 0x0002;
      break;
    default   : return (-1);
  }

  RS232_Addr = MK_FP(0x0040, Offset);  /* Find out where the port is. */
  if (*RS232_Addr == NULL) return (-1);/* If NULL then port not used. */
    portbase = *RS232_Addr;              /* Otherwise set portbase      */

  return (0);
}

/* This routine sets the speed; will accept funny baud rates. */
/* Setting the speed requires that the DLAB be set on.        */
int SetSpeed(int Speed)
{
  char c;
  int  divisor;

  if (Speed == 0)            /* Avoid divide by zero */
    return (-1);
  else
    divisor = (int) (115200L/Speed);

  if (portbase == 0)
    return (-1);

  disable();
  c = inportb(portbase + LCR);
  outportb(portbase + LCR, (c | 0x80)); /* Set DLAB */
  outportb(portbase + DLL, (divisor & 0x00FF));
  outportb(portbase + DLH, ((divisor >> 8) & 0x00FF));
  outportb(portbase + LCR, c);          /* Reset DLAB */
  enable();

  return (0);
}

/* Set other communications parameters */
int    SetOthers(int Parity, int Bits, int StopBit)
{
    int                setting;

    if (portbase == 0)                  return (-1);
    if (Bits < 5 || Bits > 8)               return (-1);
    if (StopBit != 1 && StopBit != 2)           return (-1);
    if (Parity != NO_PARITY && Parity != ODD_PARITY && Parity != EVEN_PARITY)
                            return (-1);

    setting  = Bits-5;
    setting |= ((StopBit == 1) ? 0x00 : 0x04);
    setting |= Parity;

    disable();
    outportb(portbase + LCR, setting);
    enable();

    return (0);
}

/* Set up the port */
int    SetSerial(int Port, int Speed, int Parity, int Bits, int StopBit)
{
    if (SetPort(Port))                    return (-1);
    if (SetSpeed(Speed))                  return (-1);
    if (SetOthers(Parity, Bits, StopBit)) return (-1);

    return (0);
}

/*  Control-Break interrupt handler */
int c_break(void)
{
  i_disable();
  closeserial();
  return(0);
}

