/* program : pc2key.c
   author  : Ren Bodenstaff
   date    : august 14 1994
   purpose : swaps member groups for key lock
   tested  : <none>
   nb      : compile this program with Test Stack Overflow OFF
*/

/* ---------------- Includes ------------------------------ */
#include <time.h>
#include <stdlib.h>
#include <io.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "serial.h"

/* ---------------- Defines ------------------------------- */
#define VERSION 0x00

#define FALSE           0
#define TRUE           (!FALSE)

#define NOERROR         0       /* no error               */
#define BUFOVFL         1       /* buffer overflowed      */

#define CR		          0x0d	  /* ascii cr*/
#define ESC             0x1b    /* ascii escape character */
#define ASCII           0x007f  /* mask ASCII characters  */
#define SBUFSIZ         0x20    /* serial buffer size     */

#define MAXLINE         80
#define ID_LEN          7       /* length of member number + '\0' in bytes */
#define MAX_MEM         680     /* maximum number of members per lock      */
#define LOCK_LEN	      9	      /* length of lock id in bytes + '\0' */

/* ---------------- Prototypes ---------------------------- */

/* --- own functions */
void scan_member(char *str, char *mem);
int read_names(char *list[], FILE *fp);
int read_lock(char *list[]);
void dump_list(char *list[], int count);
void write_list(char *list[], int count);
int create_union_list(char *cur_list[], char *mem_list[], char *uni_list[],
		      int cur_count, int mem_count);
int create_add_list(char *mem_list[], char *uni_list[], char *add_list[],
		    int mem_count, int uni_count);
int create_del_list(char *cur_list[], char *uni_list[], char *del_list[],
		    int cur_count, int uni_count);
int is_element(char *str, char *list[], int count);
void del_key(char *list[], int count);
void add_key(char *list[], int count);
int soft_char(char ch);
void flush_buf(void);

/* --- serial functions */
void interrupt com_int(void);
int SerialOut(char x);
int SerialOutTough(char x);
void SerialString(char *string);
int getccb(void);
void setvects(void);
void resvects(void);
void i_enable(int pnum);
void i_disable(void);
void comm_on(void);
void comm_off(void);
void initserial(void);
void closeserial(void);
int SetPort(int Port);
int SetSpeed(int Speed);
int SetOthers(int Parity, int Bits, int StopBit);
int SetSerial(int Port, int Speed, int Parity, int Bits, int StopBit);
int c_break(void);

/* -------------------------- MAIN ---------------------------- */

/* --- some global variables */
  int SError = NOERROR;
  int portbase = 0;
  void interrupt(*oldvects[2])();

  static char ccbuf[SBUFSIZ];
  unsigned int startbuf = 0;
  unsigned int endbuf = 0;

main(int argc, char *argv[])
{
  FILE *names_file;
  int i, n, port;
  unsigned member_count;
  unsigned current_count;
  unsigned union_count;
  unsigned add_count;
  unsigned del_count;
  char *member_list[MAX_MEM];
  char *current_list[MAX_MEM];
  char *union_list[MAX_MEM];
  char *add_list[MAX_MEM];
  char *del_list[MAX_MEM];
  char ch;
  char line[MAXLINE];

  int c;

  int speed    = 9600;
  int parity   = NO_PARITY;
  int bits     = 8;
  int stopbits = 1;

  /* --- check for command line consistency & open names file and scan port# */
  if(argc!=3 || (strcmp(argv[2], "com1") && strcmp(argv[2], "com2") &&
		 strcmp(argv[2], "COM1") && strcmp(argv[2], "COM2"))) {
    fprintf(stderr, "usage: pc2key <names_file> <com1|com2>\n");
    exit(1);
  }
  if((names_file = fopen(argv[1], "rt")) == NULL) {
    fprintf(stderr, "can't open %s\n", argv[1]);
    exit(1);
  }
  else {
    fprintf(stdout, "file %s opened for reading\n", argv[1]);

    if(strcmp(argv[2], "com1"))
      port=COM2;
    else
      port=COM1;
  }

  /* --- first set communications parameters */

  if (SetSerial(port, speed, parity, bits, stopbits) != 0) {
    fprintf(stderr, "Serial Port setup error.\n");
    return (99);
  }

  initserial();
  ctrlbrk(c_break);

  /* --- build lists of current and member numbers */
  member_count=read_names(member_list, names_file);
  current_count=read_lock(current_list);

  /* --- save current numbers to disk */
  write_list(current_list, current_count);

  /* --- derive list for adding and deleting members */
  union_count=create_union_list(current_list, member_list, union_list,
				current_count, member_count);


  add_count=create_add_list(member_list, union_list, add_list,
			    member_count, union_count);


  del_count=create_del_list(current_list, union_list, del_list,
			    current_count, union_count);

  /* --- confuse user :) */
  fprintf(stdout, "member list entries  : %d\n", member_count);
  fprintf(stdout, "current list entries : %d\n", current_count);
  fprintf(stdout, "union list entries   : %d\n", union_count);
  fprintf(stdout, "add list entries     : %d\n", add_count);
  fprintf(stdout, "del list entries     : %d\n", del_count);

  /* --- check whether master key is member of del_list */
  if(is_element(current_list[0], del_list, del_count)) {
    fprintf(stderr, "fatal error: attempt to delete master key, \
		     program aborted\n");
    exit(1);
  }
  else { /* --- update membership of lock */
    if(del_count)
      del_key(del_list, del_count);

    if(add_count)
      add_key(add_list, add_count);
  }

  /* --- make sure lock is _not_ in host mode (rude way) */
  SerialOutTough('r');

  /* --- check for serial errors & close down */
  switch (SError)
  {
    case NOERROR: closeserial();
                  return(0);
    case BUFOVFL: /* fprintf(stderr, "\nBuffer Overflow\n"); */
                  closeserial();
                  return (99);

    default:      fprintf(stderr, "\nUnknown Error, SError = %d\n", SError);
                  closeserial();
                  return (99);
  } /* switch */
} /* main */

/* ---------------- Functions ---------------------------------- */

/* --- functions for list data structure */

void scan_member(char *str, char *mem)
{
  int i;

  /* scan str[] for member number and store in mem[] */
  for(i=0; isalnum(str[i]); mem[i]=str[i], i++);
    mem[i]='\0';
} /* scan_member */


int read_names(char *list[], FILE *fp)
{
  /* --- read names file and store member numbers in member_list
     --- while keeping track of number of members
  */

  char ch, line[MAXLINE], member_number[ID_LEN];
  char *tmp;
  int i=0;
  int count=0;

  while((ch=(char)fgetc(fp)) != EOF) {
    if(ch=='#')
      while((ch=(char)fgetc(fp)) != '\n');
    else {
      if(ch!='\n')
	line[i++]=ch;
      else {
	line[i]='\0';
	scan_member(line, member_number);
	if((tmp=malloc(ID_LEN)) == NULL) {
	  fprintf(stderr, "r_n: out of memory\n");
	  return(-1);
	} /* if */
	else {
	strcpy(tmp, member_number);
	  if(strlen(tmp))
	    list[count++]=tmp;
	  i=0;
	} /* else */
      } /* else */
    } /* else */
  } /* while != EOF */

  fclose(fp);
  return (count);
} /* read_names */


int read_lock(char *list[])
{
  /* --- read key members from lock by asserting 's' command
     --- and store current member numbers in lock_list
     --- nb: beware of master number (never delete this one)!!
  */

  char ch, line[MAXLINE], lock_id[LOCK_LEN];
  char *tmp;
  int i=0;
  int count=0;

  /* --- send '\n' to get hold of the lock's id */
  SerialOutTough('?');

  i=0;
  do {
    if((ch=(char)getccb())!=-1) {
      switch(ch) {
	case '0': case '1': case '2': case '3': case '4': case '5':
	case '6': case '7': case '8': case '9': case 'a': case 'b':
	case 'c': case 'd': case 'e': case 'f':
		  lock_id[i++]=ch;
		  break;
	default: break;
      }
    } /* if((ch=(char)....) */
  } while (i<LOCK_LEN-1);
  lock_id[i]='\0';

  fprintf(stdout, "lock id %s\n", lock_id);

  /* --- delay for safety (lock's kinda slow....) */
  delay(500);

  /* --- send 's' command and collect data in list */
  SerialOutTough('s');

  i=0;
  do {
  if((ch=(char)getccb())!=-1) {
    if(ch!='\n')
      line[i++]=ch;
    else {
      if(!(i<ID_LEN)) { /* don't store non-valid lines */
	line[i-1]='\0';
	if((tmp=malloc(ID_LEN)) == NULL) {
	  fprintf(stderr, "r_l: out of memory\n");
	  return(-1);
	} /* malloc */
	else {
	  strcpy(tmp, line);
	  list[count++]=tmp;
	}
      }
      i=0;
    }
  } /* if((ch=(char)....) */
  } while (i<LOCK_LEN);

  return(count);
} /* read_lock */


void dump_list(char *list[], int count)
{ /* dumps list to screen */
  int i;
  for(i=0; i<count; i++)
    fprintf(stdout, "d_l: count: %d, id number: %s\n", i, list[i]);
} /* dump_list */


void write_list(char *list[], int count)
{ /* writes list to standard file <keylist.txt> */
  int i;
  FILE *key_list;

  if((key_list = fopen("keylist.txt", "wt")) == NULL) {
    fprintf(stderr, "can't create keylist.txt\n");
    exit(1);
  }

  for(i=0; i<count; i++)
    fprintf(key_list, "%s\n", list[i]);

  fclose(key_list);
} /* write_list */


int create_union_list(char *cur_list[], char *mem_list[], char *uni_list[],
		      int cur_count, int mem_count)
{ /* --- derive list of id's that is the union of the member_list
     --- and the current_list
  */
  int i, j;
  int count=0;
  char *tmp;

  for(i=0; i<cur_count; i++) {
    if((is_element(cur_list[i], mem_list, mem_count))) {
	if((tmp=malloc(ID_LEN)) == NULL) {
	  fprintf(stderr, "c_a_l: out of memory\n");
	  return(-1);
	} /* malloc */
	else {
	  strcpy(tmp, cur_list[i]);
	  uni_list[count++]=tmp;
	} /* else */
      } /* if */
  } /* for */

  return(count);
} /* create_union_list */


int create_add_list(char *mem_list[], char *uni_list[], char *add_list[],
		    int mem_count, int uni_count)
{ /* --- derive list of members to be added, that is:
     --- member_list minus union
  */
  int i, j;
  int count=0;
  char *tmp;

  for(i=0; i<mem_count; i++) {
    if(!(is_element(mem_list[i], uni_list, uni_count))) {
	if((tmp=malloc(ID_LEN)) == NULL) {
	  fprintf(stderr, "c_a_l: out of memory\n");
	  return(-1);
	} /* malloc */
	else {
	  strcpy(tmp, mem_list[i]);
	  add_list[count++]=tmp;
	} /* else */
      } /* if */
  } /* for */

  return(count);
} /* create_add_list */


int create_del_list(char *cur_list[], char *uni_list[], char *del_list[],
		     int cur_count, int uni_count)
{ /* --- derive list of members to be deleted, that is:
     --- current_list minus union
  */
  int i, j;
  int count=0;
  char *tmp;

  for(i=1; i<cur_count; i++) { /* i=1: never deleted master !! */
    if(!(is_element(cur_list[i], uni_list, uni_count))) {
	    if((tmp=malloc(ID_LEN)) == NULL) {
	      fprintf(stderr, "c_a_l: out of memory\n");
	      return(-1);
      } /* malloc */
	    else {
      strcpy(tmp, cur_list[i]);
	    del_list[count++]=tmp;
	    } /* else */
    } /* if */
  } /* for */
  return(count);
} /* create_del_list */

int is_element(char *str, char *list[], int count)
{ /* element=(str is in list) */
  int i;

  for(i=0; i<count; i++)
    if(strcmp(str, list[i]) == 0)
      return (1);

  return 0;
} /* is_element */


void del_key(char *list[], int count)
{ /* --- update membership of lock with 'd' command */
  int c=0, n, num=0, err=0;
  char ch, *id;

  while(c<count) {
    n=0;
    id=list[c];

    if(soft_char('d'))
      err=1;
    while(n<ID_LEN-1 && !err) {
      if(soft_char(id[n]))
        err=1;
      n++;
    }
    if(!err)
      if(soft_char(0x0d))
        err=1;

    if(!err) {
      do {
        if((ch=(char)getccb())!=-1) {
	        fprintf(stdout, "%c", ch);
	        num++;
        }
      } while(num<11);
      num=0;
    } /* if */

    if(!err)
      c++;
    err=0;
  } /* while */
} /* del_key */


void add_key(char *list[], int count)
{ /* --- update membership of lock with 'n' command */
  int c=0, n, num=0, err=0;
  char ch, *id;

  while(c<count) {
    n=0;
    id=list[c];

    if(soft_char('n'))
      err=1;
    while(n<ID_LEN-1 && !err) {
      if(soft_char(id[n]))
        err=1;
      n++;
    }
    if(!err)
      if(soft_char(0x0d))
        err=1;

    if(!err) {
      do {
        if((ch=(char)getccb())!=-1) {
	        fprintf(stdout, "%c", ch);
	        num++;
        }
      } while(num<11);
      num=0;
    } /* if */

    if(!err)
      c++;
    err=0;
  } /* while */
} /* add_key */


int soft_char(char ch)
{ /* send char to serial port and wait for echo */
  char echo;

  SerialOutTough(ch);

  while((echo=(char)getccb())==-1) {}
    fprintf(stdout, "%c", echo);
  if(ch==0x0d && echo=='!')
    return 0;

  if(ch!=echo) {
    flush_buf();
    return 1;
  }
  return 0;
}

void flush_buf(void)
{ /* flushes buffer when out of sync */
  char ch;

  /* reset lock & flush buffer (rude way) */
  SerialOutTough(0x0d);
  delay(250);
  startbuf=0;
  endbuf=0;
} /* flush_buf */

/* --- functions for RS-232 communications (ripped from serial.c) */

/* Handle communications interrupts and put them in ccbuf */
void interrupt com_int(void)
{
  disable();
  if ((inportb(portbase + IIR) & RX_MASK) == RX_ID) {
    if (((endbuf + 1) & SBUFSIZ - 1) == startbuf)
      SError = BUFOVFL;

    ccbuf[endbuf++] = inportb(portbase + RXR);
    endbuf &= SBUFSIZ - 1;
  }

  /* Signal end of hardware interrupt */
  outportb(ICR, EOI);
  enable();
}

/* Output a character to the serial port */
int SerialOut(char x)
{
  long int timeout = 0x0000FFFFL;

  outportb(portbase + MCR,  MC_INT | DTR | RTS);

  /* Wait for Clear To Send from modem */
  while ((inportb(portbase + MSR) & CTS) == 0)
    if (!(--timeout))
      return (-1);

  timeout = 0x0000FFFFL;

  /* Wait for transmitter to clear */
  while ((inportb(portbase + LSR) & XMTRDY) == 0)
    if (!(--timeout))
      return (-1);

  disable();
  outportb(portbase + TXR, x);
  enable();

  return (0);
}

/* Output a character to the serial port with no hand shaking */
/* --- This hack is necessary, because the key lacks RTS & CTS --- */
int SerialOutTough(char x)
{
  outportb(portbase + MCR,  MC_INT | DTR | RTS);

  disable();
  outportb(portbase + TXR, x);
  enable();

  return (0);
}

/* Output a string to the serial port */
void SerialString(char *string)
{
  while (*string)
    SerialOut(*string++);
}

/* This routine returns the current value in the buffer */
int getccb(void)
{
  int res;

  if (endbuf == startbuf)
    return (-1);

  res = (int) ccbuf[startbuf++];
  startbuf %= SBUFSIZ;

  return (res);
}

/* Install our functions to handle communications */
void setvects(void)
{
  oldvects[0] = getvect(0x0B);
  oldvects[1] = getvect(0x0C);
  setvect(0x0B, com_int);
  setvect(0x0C, com_int);
}

/* Uninstall our vectors before exiting the program */
void resvects(void)
{
  setvect(0x0B, oldvects[0]);
  setvect(0x0C, oldvects[1]);
}

/* Turn on communications interrupts */
void i_enable(int pnum)
{
  int c;

  disable();
  c = inportb(portbase + MCR) | MC_INT;
  outportb(portbase + MCR, c);
  outportb(portbase + IER, RX_INT);
  c = inportb(IMR) & (pnum == COM1 ? IRQ4 : IRQ3);
  outportb(IMR, c);
  enable();
}

/* Turn off communications interrupts */
void i_disable(void)
{
  int c;

  disable();
  c = inportb(IMR) | ~IRQ3 | ~IRQ4;
  outportb(IMR, c);
  outportb(portbase + IER, 0);
  c = inportb(portbase + MCR) & ~MC_INT;
  outportb(portbase + MCR, c);
  enable();
}

/* Tell modem that we're ready to go */
void comm_on(void)
{
  int c, pnum;

  pnum = (portbase == COM1BASE ? COM1 : COM2);
  i_enable(pnum);
  c = inportb(portbase + MCR) | DTR | RTS;
  outportb(portbase + MCR, c);
}

/* Go off-line */
void comm_off(void)
{
  i_disable();
  outportb(portbase + MCR, 0);
}

void initserial(void)
{
  endbuf = startbuf = 0;
  setvects();
  comm_on();
}

void closeserial(void)
{
  comm_off();
  resvects();
}

/* Set the port number to use */
int SetPort(int Port)
{
  int  Offset, far *RS232_Addr;

  switch (Port)
  { /* Sort out the base address */
    case COM1 : Offset = 0x0000;
      break;
    case COM2 : Offset = 0x0002;
      break;
    default   : return (-1);
  }

  RS232_Addr = MK_FP(0x0040, Offset);  /* Find out where the port is. */
  if (*RS232_Addr == NULL) return (-1);/* If NULL then port not used. */
    portbase = *RS232_Addr;              /* Otherwise set portbase      */

  return (0);
}

/* This routine sets the speed; will accept funny baud rates. */
/* Setting the speed requires that the DLAB be set on.        */
int SetSpeed(int Speed)
{
  char c;
  int  divisor;

  if (Speed == 0)            /* Avoid divide by zero */
    return (-1);
  else
    divisor = (int) (115200L/Speed);

  if (portbase == 0)
    return (-1);

  disable();
  c = inportb(portbase + LCR);
  outportb(portbase + LCR, (c | 0x80)); /* Set DLAB */
  outportb(portbase + DLL, (divisor & 0x00FF));
  outportb(portbase + DLH, ((divisor >> 8) & 0x00FF));
  outportb(portbase + LCR, c);          /* Reset DLAB */
  enable();

  return (0);
}

/* Set other communications parameters */
int    SetOthers(int Parity, int Bits, int StopBit)
{
    int                setting;

    if (portbase == 0)                  return (-1);
    if (Bits < 5 || Bits > 8)               return (-1);
    if (StopBit != 1 && StopBit != 2)           return (-1);
    if (Parity != NO_PARITY && Parity != ODD_PARITY && Parity != EVEN_PARITY)
                            return (-1);

    setting  = Bits-5;
    setting |= ((StopBit == 1) ? 0x00 : 0x04);
    setting |= Parity;

    disable();
    outportb(portbase + LCR, setting);
    enable();

    return (0);
}

/* Set up the port */
int    SetSerial(int Port, int Speed, int Parity, int Bits, int StopBit)
{
    if (SetPort(Port))                    return (-1);
    if (SetSpeed(Speed))                  return (-1);
    if (SetOthers(Parity, Bits, StopBit)) return (-1);

    return (0);
}

/*  Control-Break interrupt handler */
int c_break(void)
{
  i_disable();
  closeserial();
  return(0);
}

